!******************************************************************************!
!*                                  META.CLW                                  *!
!*                                                                            *!
!*      Copyright (C) 1995 P.S. Wilson & Associates, all rights reserved.     *!
!*                                                                            *!
!*  Meta File Driver: This is a limited functionality Meta Driver which may   *!
!*  be used to dynamically load any driver for a given file structure.        *!
!*                                                                            *!
!*  IMPORTANT: This code may be used by anyone for any purpose in it's        *!
!*  original or modified states provided that all copyright notices remain    *!
!*  intact, I reserve the right to change this status for future versions.    *!
!*                                                                            *!
!*  DISCLAIMER: I can accept no responsibility whatsoever for any damages of  *!
!*  any kind arising from the use or misuse of this or any associated code or *!
!*  information. You are considered entirely responsible for your own actions *!
!*  and decisions.                                                            *!
!*                                                                            *!
!*  Sean Wilson, 29 March 1996                                                *!
!******************************************************************************!
MetaPipe        PROGRAM

                INCLUDE('Windows.INC')

!*** Clarion for Windows data types ********************************************
StrDesc         GROUP, TYPE                 ! String descriptor
Str               POINTER                   ! String data pointer
Len               UNSIGNED                  ! String data length
                END

ClaFCB          GROUP, TYPE                 ! File Control Block Structure
Pipe              FARPROC                   ! Pointer to pipe function
PhyName           LIKE(StrDesc)             ! Physical filename
DrvStr            LIKE(StrDesc)             ! File driver string
OwnerName         LIKE(StrDesc)             ! Owner name for file
FileAttr          BYTE                      ! File attributes
KeyCount          BYTE                      ! Number of keys in this file
MemoCount         BYTE                      ! Number of memos in this file
RecordLen         UNSIGNED                  ! Record length
TDesc             POINTER                   ! Pointer to RECORD type desc
KDesc             POINTER                   ! Pointer to KEY descriptor array
MDesc             POINTER                   ! Pointer to MEMO descriptor array
RecordBuf         POINTER                   ! Pointer to record buffer
RBlock            POINTER                   ! File drivers runtime block
TID               BYTE                      ! Owner thread
TFlags            BYTE                      ! Owner flags
                END

!*** Meta Driver data types ****************************************************
MetaFCB         QUEUE                       ! Meta Driver's runtime data
drvFcb            POINTER                   ! Driver's Fcb (used as a key)
drvHnd            HINSTANCE                 ! Driver's module handle
drvPipe           FARPROC                   ! Driver's pipe function
                END

!*** Meta Driver procedure map *************************************************
                MAP
                  Meta(SIGNED, POINTER, POINTER), LONG
                  MetaShutdown()
                  GetDriver(ClaFCB, *CSTRING)
                  InitMcb(*ClaFCB)
                  MetaError(STRING)

                  MODULE('MetaThunk')
                    CallPipe(FARPROC, SIGNED, POINTER, POINTER), LONG
                  END

                  MODULE('RTL')
                    StrNCpy(*CSTRING, POINTER, UNSIGNED), CSTRING, RAW, PROC, NAME('_strncpy')
                    MemCpy(POINTER, POINTER, UNSIGNED), POINTER, PROC, NAME('_memcpy')
                  END
                END

  CODE


!*** METAERROR *****************************************************************
!*
!*  Synopsis:
!*      MetaError(STRING s)
!*
!*  Description:
!*      MetaError() is called to handle Meta Driver errors, since these are all
!*      unrecoverable it simply displays an error message then halts.
!*
!*  NOTE:
!*      You can use SHUTDOWN() to register a handler to shut down parts of the
!*      application gracefully.

MetaError       PROCEDURE(s)
  CODE
COMPILE('Win32', _WIN32)
  s = s & ' (' & GetLastError() & ')'
Win32
  MESSAGE(s, 'Meta Driver Error', ICON:Exclamation, BUTTON:OK)
  !*** Copyright data, please leave!
  s = 'Meta Driver, Copyright (C) 1996 P.S. Wilson & Associates'
  HALT


!*** GETDRIVER *****************************************************************
!*
!*  Synopsis:
!*      GetDriver(*ClaFCB Fcb, *CSTRING Path)
!*
!*  Description:
!*      GetDriver() parses the file's driver string to find a driver name.  This
!*      must be of the form: /DRIVER=CWCLA32.DLL.  There should be no spaces in
!*      the string.  The name should name the driver .DLL.
!*
!*  NOTE:

GetDriver       PROCEDURE(Fcb, Path)
Str               CSTRING(256), AUTO
  CODE
  StrNCpy(Str, Fcb:DrvStr:Str, SIZE(Str))

  i# = INSTRING('/DRIVER=', Str, 1)

  IF i# <> 0 THEN
    j# = INSTRING(' ', Str, 1, i#+1)

    IF j# <> 0 THEN
      Path = Str[i#+8 : i#+j#-1]
    ELSE
      Path = Str[i#+8 : LEN(Str)]
    END
  END

  IF Path = '' THEN
    MetaError('No driver specified')
  END


!*** INITMCB *******************************************************************
!*
!*  Synopsis:
!*      InitMcb(*ClaFCB Fcb)
!*
!*  Description:
!*      InitMcb() is called when there is no entry in the MetaFCB queue to match
!*      the current file.  It obtains the file's driver name, dynamically loads
!*      the driver and determines it's entry point.  This inofrmation will be
!*      stored in the MetaFCB queue.
!*
!*  NOTE:

InitMcb         PROCEDURE(Fcb)
DrvPath           CSTRING(256), AUTO
  CODE
  GetDriver(Fcb, DrvPath)
  MetaFCB:drvHnd = LoadLibrary(drvPath)

  IF MetaFCB:drvHnd > HINSTANCE_ERROR THEN
    MetaFCB:drvPipe = GetOrdAddress(MetaFCB:drvHnd, 2)

    IF MetaFCB:drvPipe = 0 THEN
      MetaError('Can''t get driver entry point')
    END
  ELSE
    MetaError('Can''t load driver')
  END


!*** META **********************************************************************
!*
!*  Synopsis:
!*      Meta(SIGNED Op, POINTER FcbPtr, POINTER Args)
!*
!*  Description:
!*      Meta() is a replacement driver entry point.  It is called to process any
!*      file command for a file defined in the following way:
!*
!*      f   FILE,DRIVER('Meta', '/DRIVER=CWTPS32.DLL')
!*              ...
!*          END
!*
!*      It processes the call by determining which driver has been specified to
!*      handle the call.  It then calls the driver supplying the received
!*      arguments.
!*
!*  NOTE:
!*      This function has been coded with robustness in mind rather than
!*      efficiency and I may have erred slightly on the side of caution.  See
!*      the comments for a detailed explanation.

META            FUNCTION(Op, FcbPtr, Args)
Fcb               LIKE(ClaFCB), AUTO
Ret               LONG, AUTO
  CODE
  !*** Locate the Meta Driver's control data from the MetaFCB queue.  NOTE
  !*** that entries are keyed on the address of the file descriptor.
  MetaFCB:drvFcb = FcbPtr
  GET(MetaFCB, MetaFCB:drvFcb)

  !*** If a matching entry doesn't exist then we need to load the Fcb so that
  !*** we can create an entry to be added to the MetaFCB queue.
  IF ERRORCODE() = 30 THEN
    MemCpy(ADDRESS(Fcb), FcbPtr, SIZE(ClaFCB))
    InitMcb(Fcb)
    ADD(MetaFCB, MetaFcb:drvFcb)

    IF ERRORCODE() THEN
      MetaError('Can''t create control data entry')
    END
  END

  !*** Now we call the requested driver.  This is done through a 'thunk'.
  !*** See the comments in the MetaThunk module.
  Ret = CallPipe(MetaFCB:drvPipe, Op, FcbPtr, Args)

  !*** Now we load the Fcb to check the drivers runtime data.  This will
  !*** be 0 if the file has been closed.  This is our cue to delete the
  !*** queue data for this file.
  MemCpy(ADDRESS(Fcb), FcbPtr, SIZE(ClaFCB))

  IF Fcb:Rblock = 0 THEN
    !*** Reload the MetaFCB data.  I do it this way so that it should be
    !*** proofed against mdifications by you.  Yes you!
    MetaFCB:drvFcb = FcbPtr
    GET(MetaFCB, MetaFCB:drvFcb)

    IF NOT ERRORCODE() THEN
      FreeModule(MetaFCB:drvHnd)
      DELETE(MetaFCB)
    END
  END

  RETURN Ret


!*** METASHUTDOWN **************************************************************
!*
!*  Synopsis:
!*      MetaShutdown()
!*
!*  Description:
!*      MetaShutdown() should be called to free all resources used by the Meta
!*      driver.  A good place to call this is from the procedure named as the
!*      argument for the SHUTDOWN() procedure.
!*
!*  NOTE:

MetaShutdown    PROCEDURE
  CODE
  LOOP
    GET(MetaFCB, 1)
    IF ERRORCODE() THEN
      BREAK
    END
    FreeModule(MetaFCB:drvHnd)
    DELETE(MetaFCB)
  END
